#******************************************************************************
# ESO/DFS
#
# "@(#) $Id: __init__.py,v 1.186 2005/10/17 09:56:08 jknudstr Exp $"
#
# Who       When        What
# --------  ----------  -------------------------------------------------------
# jknudstr  12/04/2001  Created
# awicenec  29/05/2001  Added path extension
# jknudstr  11/06/2001  Added proper version + implemented getNgamsVersion()

_doc =\
"""
             #     #  #####        #    #    #     #  #####
             ##    # #     #      #    # #   ##   ## #     #
             # #   # #           #    #   #  # # # # #
             #  #  # #  ####    #    #     # #  #  #  #####
             #   # # #     #   #     ####### #     #       #
             #    ## #     #  #      #     # #     # #     #
             #     #  #####  #       #     # #     #  #####


%s


The NG/AMS SW provides a server to be used on the NGAS System for
handling the archival and retrieval of data, and to manage the
disk configuration of the system.

The complete functionality is implemented as a set of Python classes
that can be used in other contexts. 

Also provided are functions used to built the so-called Data Handling
Plug-In Utilities which are used to handle the various types of data
files being archived.

For more detailed information about the project, consult the URL:

  http://www.eso.org/projects/dfs/dfs-shared/web/ngas/ngas-general.html


To use the NG/AMS classes, the following statement should be contained
in the Python source files:

from   ngams import *

Subsequently the functions and variables defined in this module can be used
without referring to the module. The NG/AMS modules can be imported in the
standard way, e.g.:

import ngamsLib, ngamsConfig, ngamsHttpPars


HIERARCHY/STRUCTURE OF THE NG/AMS SOURCE CODE FILES:
The following diagram indicates how the hierarchy is structured for the Python
source code files contained within NG/AMS. The files (Python modules) should
only be imported by other modules, which have a lower location in the diagram,
never opposite, nor by files at the same level. This in principle is allowed by
Python, but in general such circular dependencies should be avoided in order to
keep a clear structure/architecture of the SW for maintability reasons.

Source Code File Hierarchy:

  o ngamsDb, ngamsLib, ngamsUrlLib
  o ngamsConfig
  o ngamsFileInfo, ngamsHostInfo
  o ngamsDiskInfo, ngamsDapiStatus, ngamsDppiStatus, ngamsFileList, 
    ngamsPhysDiskInfo
  o ngamsReqProps, ngamsStatus
  o ngamsHighLevelLib
  o ngamsDiskUtils
  o ngamsPlugInApi

The source files contained in the sub-modules 'ngamsServer', 'ngamsPClient',
'ngamsTest' and 'ngamsCClient' are not considered in the list above. The
sub-module contaning example plug-ins 'ngamsPlugIns' is not considered either.
"""


# Debug flag.
_debug = 0

# Flag indicating if we're executing in Unit Test Mode.
_testMode = 0


# Make the NG/AMS classes available by extending the search paths.
try:
    from sys import path
    pathTup = [__path__[0] + '/../plug-ins',
               __path__[0] + '/ngamsCClient',
               __path__[0] + '/ngamsData',
               __path__[0] + '/ngamsLib',
               __path__[0] + '/ngamsPClient',
               __path__[0] + '/ngamsPlugIns',
               __path__[0] + '/ngamsServer',
               __path__[0] + '/ngamsTest']
    path.extend(pathTup)
    __path__.extend(pathTup)
except Exception, e:
    pass

import os, string, re, syslog, traceback, threading, types, time, commands
import pcc, PccUtString, PccLogDef, PccLog, PccUtTime


# NG/AMS source directory
NGAMS_SRC_DIR = os.path.dirname(__file__)


# Main PID of server
NGAMS_SRV_PID = os.getpid()


# Semaphore + counter to ensure unique, temporary filenames.
_uniqueNumberSem   = threading.Semaphore(1)
_uniqueNumberCount = 0 


def ngamsGetSrcDir():
    """
    Return the NG/AMS source directory, i.e., the directory where
    the modules of NG/AMS are contained.

    Returns:  NG/AMS source directory (string).
    """
    global NGAMS_SRC_DIR
    return NGAMS_SRC_DIR


def getSrvPid():
    """
    Get the main PID of the Python interpreter, in which the NG/AMS Server
    or another application based on the NG/AMS library, is running.
    
    Returns:    Main PID of server (integer).
    """
    global NGAMS_SRV_PID
    return NGAMS_SRV_PID


# Import COPYRIGHT statement into doc page.
fo = open(os.path.normpath(ngamsGetSrcDir() + "/COPYRIGHT"))
NGAMS_COPYRIGHT_TEXT = fo.read()
fo.close()
__doc__ = _doc % NGAMS_COPYRIGHT_TEXT


# Handle NG/AMS Version.
fo = open(os.path.normpath(ngamsGetSrcDir() + "/VERSION"))
verBufLines = fo.readlines()
fo.close()
_NGAMS_CVS_ID   = "--UNDEFINED--"
_NGAMS_SW_VER   = "--UNDEFINED--"
_NGAMS_VER_DATE = "--UNDEFINED--"
for line in verBufLines:
    if (line.find("NGAMS_CVS_ID") != -1):
        _NGAMS_CVS_ID = line.split("NGAMS_CVS_ID ")[1].strip()[1:-1]
    elif (line.find("NGAMS_SW_VER") != -1):
        _NGAMS_SW_VER = line.split("NGAMS_SW_VER ")[1].strip()[1:-1]
    elif (line.find("VER_DATE") != -1):
        _NGAMS_VER_DATE = line.split("VER_DATE ")[1].strip()[1:-1]
    elif ((_NGAMS_CVS_ID != "") and (_NGAMS_SW_VER != "") and
          (_NGAMS_VER_DATE != "")):
        break


# Load Error Definition File
NGAMS_ERR_DEF_FILE = os.path.normpath(ngamsGetSrcDir() + "/ngamsData/" +\
                                      "ngamsLogDef.xml")
_logDef = PccLogDef.PccLogDef().load(NGAMS_ERR_DEF_FILE)


# Flag used to suppress error logging on stderr.
_suppresErrorLogging = 0


# Variable indicating port number used in case multiple servers are executed
# on the same node. In this case, the Host ID used for addressing will be
# <Host Name>:<Port No> for unique addressing.
# If this feature is not used, this should be None.
_srvPortNo = None


# Log protection semaphore.
_logSem = threading.Semaphore(1)


def getNgamsLicense():
    """
    Read in and return the NG/AMS License Agreement.

    Returns:   Contents of license agreement (string).
    """
    fo = open(os.path.normpath(ngamsGetSrcDir() + "/LICENSE"))
    license = fo.read()
    fo.close()
    return license


def prFormat1():
    """
    Return format used when dumping contents of NG/AMS objects.

    Return:    Format (string).
    """
    return "%-35s %s\n"

_CONST_FORMAT = "\n%-32s = %s"

# Directories and Filenames.
NGAMS_BAD_FILES_DIR           = "bad-files"
NGAMS_BACK_LOG_DIR            = "back-log"
NGAMS_SUBSCR_BACK_LOG_DIR     = "subscr-back-log"
NGAMS_SUBSCR_BACK_LOG         = NGAMS_SUBSCR_BACK_LOG_DIR
NGAMS_STAGING_DIR             = "staging"
NGAMS_PROC_DIR                = "processing"
NGAMS_FILE_DB_COUNTER         = "__COUNT__"
NGAMS_TMP_FILE_PREFIX         = "NGAMS_TMP_FILE___"
NGAMS_BACK_LOG_TMP_PREFIX     = "NGAMS_BACK_LOG_TMP___"
NGAMS_BAD_FILE_PREFIX         = "BAD-FILE-"
NGAMS_PICKLE_FILE_EXT         = "pickle"

# Status.
NGAMS_FAILURE             = "FAILURE"
NGAMS_SUCCESS             = "SUCCESS"

# States.
NGAMS_OFFLINE_STATE       = "OFFLINE"
NGAMS_ONLINE_STATE        = "ONLINE"
NGAMS_NOT_RUN_STATE       = "NOT-RUNNING"

# Sub-States.
NGAMS_IDLE_SUBSTATE       = "IDLE"
NGAMS_BUSY_SUBSTATE       = "BUSY"

# Built-In Mime-types.
NGAMS_ARCH_REQ_MT         = "ngas/archive-request"
NGAMS_TEXT_MT             = "text/plain"
NGAMS_XML_MT              = "text/xml"
NGAMS_UNKNOWN_MT          = "unknown"

# HTTP Methods, etc.
NGAMS_HTTP_GET            = "GET"
NGAMS_HTTP_POST           = "POST"
NGAMS_HTTP_FILE_URL       = "file:"
NGAMS_HTTP_INT_AUTH_USER  = "ngas-int"

# HTTP Status Codes.
NGAMS_HTTP_SUCCESS        = 200
NGAMS_HTTP_REDIRECT       = 303
NGAMS_HTTP_BAD_REQ        = 400
NGAMS_HTTP_UNAUTH         = 401
NGAMS_HTTP_UNAUTH_STR     = "Unauthorized"

# Request Processing Data Types.
NGAMS_PROC_DATA           = "DATA"
NGAMS_PROC_FILE           = "FILE"
NGAMS_PROC_STREAM         = "STREAM"

# Commands Handled by NG/AMS:
NGAMS_ARCHIVE_CMD       = "ARCHIVE"
NGAMS_CHECKFILE_CMD     = "CHECKFILE"
NGAMS_CLONE_CMD         = "CLONE"
NGAMS_CONFIG_CMD        = "CONFIG"
NGAMS_DISCARD_CMD       = "DISCARD"
NGAMS_EXIT_CMD          = "EXIT"
NGAMS_HELP_CMD          = "HELP"
NGAMS_INIT_CMD          = "INIT"
NGAMS_LABEL_CMD         = "LABEL"
NGAMS_OFFLINE_CMD       = "OFFLINE"
NGAMS_ONLINE_CMD        = "ONLINE"
NGAMS_REGISTER_CMD      = "REGISTER"
NGAMS_REMDISK_CMD       = "REMDISK"
NGAMS_REMFILE_CMD       = "REMFILE"
NGAMS_RETRIEVE_CMD      = "RETRIEVE"
NGAMS_STATUS_CMD        = "STATUS"
NGAMS_SUBSCRIBE_CMD     = "SUBSCRIBE"
NGAMS_UNSUBSCRIBE_CMD   = "UNSUBSCRIBE"

# Types of Notification Events.
NGAMS_NOTIF_INFO        = "InfoNotification"
NGAMS_NOTIF_ALERT       = "AlertNotification"
NGAMS_NOTIF_ERROR       = "ErrorNotification"
NGAMS_NOTIF_DISK_SPACE  = "DiskSpaceNotification"
NGAMS_NOTIF_DISK_CHANGE = "DiskChangeNotification"
NGAMS_NOTIF_NO_DISKS    = "NoDiskSpaceNotification"
NGAMS_NOTIF_DATA_CHECK  = "DataCheckNotification"

# NGAS Host Location Specifiers.
NGAMS_HOST_LOCAL        = "LOCAL"
NGAMS_HOST_CLUSTER      = "CLUSTER"
NGAMS_HOST_DOMAIN       = "DOMAIN"
NGAMS_HOST_REMOTE       = "REMOTE"

# Consistency Checking:
NGAMS_FILE_STATUS_OK    = "00000000"
NGAMS_FILE_CHK_ACTIVE   = "01000000"
NGAMS_CHECK_SEQ         = "SEQUENTIAL"
NGAMS_CHECK_RAN         = "RANDOM"

# Constants for Handling XML Documents:
NGAMS_XML_STATUS_ROOT_EL= "NgamsStatus"
NGAMS_XML_STATUS_DTD    = "ngamsStatus.dtd"

# DB actions (operations that change the DB).
NGAMS_DB_DIR            = ".db"
NGAMS_DB_CH_CACHE       = NGAMS_DB_DIR + "/" + "cache"
NGAMS_CACHE_DIR         = "cache"
NGAMS_DB_SNAPSHOT       = "DB-SNAPSHOT"
NGAMS_DBM_EXT           = "bsddb"
NGAMS_DB_NGAS_FILES     = "NgasFiles.bsddb"
NGAMS_DB_LAST_DB        = "LAST-DB"
NGAMS_DB_CH_FILE_INSERT = "FILE-INSERT"
NGAMS_DB_CH_FILE_UPDATE = "FILE-UPDATE"
NGAMS_DB_CH_FILE_DELETE = "FILE-DELETE"

# Miscelleneous.
NGAMS_DISK_INFO          = "NgasDiskInfo"
NGAMS_VOLUME_ID_FILE     = ".ngas_volume_id"
NGAMS_JANITOR_THR        = "JANITOR-THREAD"
NGAMS_DATA_CHECK_THR     = "DATA-CHECK-THREAD"
NGAMS_SUBSCRIPTION_THR   = "SUBSCRIPTION-THREAD"
NGAMS_SUBSCRIBER_THR     = "SUBSCRIBER-THREAD"
NGAMS_DELIVERY_THR       = "DELIVERY-THREAD-"
NGAMS_CLONE_THR          = "CLONE-THREAD-"
NGAMS_REGISTER_THR       = "REGISTER-THREAD-"
NGAMS_DEF_LOG_PREFIX     = "NGAS-LOG-PREFIX"
NGAMS_NOT_SET            = "NOT-SET"
NGAMS_DEFINE             = "DEFINE"
NGAMS_MAX_FILENAME_LEN   = 128
NGAMS_MAX_SQL_QUERY_SZ   = 2048


def legalCmd(cmd):
    """
    Check that command given is a legal (recognized) command.

    cmd:       Command to check existence of (string).

    Returns:   1 if command is legal, 0 if not (integer).
    """
    if ((cmd == NGAMS_ARCHIVE_CMD) or (cmd == NGAMS_CLONE_CMD) or
        (cmd == NGAMS_EXIT_CMD) or (cmd == NGAMS_INIT_CMD) or
        (cmd == NGAMS_LABEL_CMD) or (cmd == NGAMS_OFFLINE_CMD) or
        (cmd == NGAMS_ONLINE_CMD) or (cmd == NGAMS_REGISTER_CMD) or
        (cmd == NGAMS_REMDISK_CMD) or (cmd == NGAMS_REMFILE_CMD) or
        (cmd == NGAMS_RETRIEVE_CMD) or (cmd == NGAMS_STATUS_CMD)):
        return 1
    else:
        return 0


def trim(str,
         trimChars):
    """
    Trim a string removing leading and trailing
    characters contained in the "trimChars" string.

    str:        String to trim (string).

    trimChars:  String containing characters to trim out of
                the input string (string).

    Returns:    Trimmed string (string).
    """
    return PccUtString.trimString(str, trimChars)


def getNgamsVersion():
    """
    Return version identifier for NG/AMS.

    Returns:  NG/AMS version ID (string).
    """
    try:
        global _NGAMS_CVS_ID, _NGAMS_SW_VER, _NGAMS_VER_DATE
        #els = string.split(_NGAMS_CVS_ID, " ")
        #date  = re.sub("/", "-", els[4])
        #time  = els[5]
        #return _NGAMS_SW_VER + "/" + date + "T" + time
        return _NGAMS_SW_VER + "/" + _NGAMS_VER_DATE
    except:
        return "--UNDFINED--"


def getNgamsVersionRaw():
    """
    Return the version of the NG/AMS SW in 'raw format'.

    Returns:    Raw version string (string).
    """
    global _NGAMS_SW_VER
    return _NGAMS_SW_VER


def ngamsCopyrightString():
    """
    Return the NG/AMS Copyright and Reference String.

    Returns:   Copyright string (string).
    """
    return NGAMS_COPYRIGHT_TEXT


def setLogCond(sysLog,
               sysLogPrefix,
               locLogLevel,
               locLogFile,
               verboseLevel):
    """
    Set the global log conditions to be used by NG/AMS.

    sysLog:         Switch logging in UNIX syslgo on/off (1/0) (integer).
    
    sysLogPrefix:   Prefix (tag) to be used in syslog entries (string).
    
    locLogLevel:    Level applied for logging into local log file (integer).
    
    locLogFile:     Name of local log file (string).
    
    verboseLevel:   Level to apply to logs written on stdout (integer).

    Note: If any of the input parameters are None, the present value will
    be maintained.
 
    Returns:        Void.
    """
    # If any of the parameters are None - take the present value.
    if (sysLog == None):
        if (PccLog.getSysLogLogLevel() == -1):
            sysLog = 0
        else:
            sysLog = PccLog.getSysLogLogLevel()
    if (sysLogPrefix == None): sysLogPrefix = PccLog.getSysLogPrefix()
    if (locLogLevel == None): locLogLevel = PccLog.getLogLevel()
    if (locLogFile == None): locLogFile = PccLog.getLogFile()
    if (verboseLevel == None): verboseLevel = PccLog.getVerboseLevel()

    # Set up the new log conditions.
    path = os.path.dirname(locLogFile)
    if ((path != "") and (not os.path.exists(path))): os.makedirs(path)
    if (sysLog == 0):
        sysLogPrio = -1
        sysLogLevel = -1
    else:
        sysLogPrio = syslog.LOG_NOTICE
        sysLogLevel = 1
    sysLogProps = [sysLogPrio, sysLogLevel, sysLogPrefix]
    PccLog.setLogCond(int(locLogLevel), locLogFile, int(verboseLevel),
                      sysLogProps, 25, 1)


def getVerboseLevel():
    """
    Get the Log Level.

    Returns:   Log Level (integer).
    """
    return PccLog.getVerboseLevel()
    

def genLog(logId,
           parList = []):
    """
    Generate a log line and return this.

    logId:    The Log ID for the log (string).

    parList:  List of parameters to fill into the log format (list).

    Returns:  Generated log line (string).
    """
    global _logDef
    for idx in range(len(parList)):
        par = parList[idx]
        if (type(par) == types.StringType):
            parList[idx] = par.replace("\n", " ")
    logMsg = _logDef.genLogX(logId, parList)
    return logMsg


def getThreadName():
    """
    Return the name of the thread or '' if this cannot be determined.

    Returns:    The name of the thread (string).
    """
    try:
        threadName = threading.currentThread().getName()
    except:
        threadName = ""
    return threadName


def getLocation():
    """
    Get location of the current position in a source file of the Python
    interpreter.

    Returns:  Location in the format: '<mod>:<method>:<ln no>' (string).
    """
    stackInfo = traceback.extract_stack()[-3]
    module = stackInfo[0].split("/")[-1]
    lineNo = stackInfo[1]
    method = stackInfo[2]
    threadName = getThreadName()
    if (threadName != ""):
        return module + ":" + method + ":" + str(lineNo) +\
               ":" + str(os.getpid()) + ":" + threadName
    else:
        return module + ":" + method + ":" + str(lineNo) +\
               ":" + str(os.getpid())


def setSuppressErrLog(state):
    """
    Set the flag indicating that error logging on stderr should be suppressed.
    Is used e.g. for the Unit Test.

    state:    State of the Suppress Error Logging Flag (integer/0|1).

    Returns:  Void.
    """
    global _suppresErrorLogging
    _suppresErrorLogging = state


def getSuppressErrLog():
    """
    Return the value of the Suppress Error Logging Flag.

    Returns:   Value of Suppress Error Logging Flag (integer/0|1).
    """
    global _suppresErrorLogging
    return _suppresErrorLogging


def takeLogSem():
    """
    Take the log protection semaphore.

    Returns:   Void.
    """
    global _logSem
    _logSem.acquire()


def relLogSem():
    """
    Release the log protection semaphore.

    Returns:  Void.
    """
    global _logSem
    _logSem.release()
   

def info(level,
         msg):
    """
    Generate an Information Log entry in the log targets.
    This is not written to UNIX syslog.
    
    level:    Level indicator for this log entry.
    
    msg:      Message to log (string).

    Returns:  Void.
    """
    try:
        takeLogSem()
        PccLog.info(level, msg, getLocation())
        relLogSem()
    except Exception, e:
        relLogSem()
        raise e


def sysLogInfo(level,
               msg,
               location = ""):
    """
    Generate a log entry in the UNIX syslog.

    msg:      Message to log (string).
    
    location: Optional location specifier (string).

    Returns:  Void.
    """
    try:
        takeLogSem()
        PccLog.sysLogInfo(level, msg, location)
        relLogSem()
    except Exception, e:
        relLogSem()
        raise e


def notice(msg):
    """
    Log a Notice Log into the specified log targets.

    msg:      Message to log.

    Returns:  Void.
    """
    if (getSuppressErrLog()): return
    try:
        takeLogSem()
        PccLog.notice(msg, getLocation())
        relLogSem()
    except Exception, e:
        relLogSem()
        raise e

    
def warning(msg):
    """
    Log a Warning Log into the specified log targets.

    msg:      Message to log.

    Returns:  Void.
    """
    if (getSuppressErrLog()): return
    try:
        takeLogSem()
        PccLog.warning(msg, getLocation())
        relLogSem()
    except Exception, e:
        relLogSem()
        raise e

    
def error(msg):
    """
    Log an Error Log into the specified log targets.

    msg:      Message to log (string).

    Returns:  Void.
    """
    if (getSuppressErrLog()): return
    try:
        takeLogSem()
        PccLog.error(msg, getLocation())
        relLogSem()
    except Exception, e:
        relLogSem()
        raise e
    

def alert(msg):
    """
    Log an Alert Log into the specified log targets.

    msg:      Message to log (string).

    Returns:  Void.
    """
    if (getSuppressErrLog()): return
    try:
        takeLogSem()
        PccLog.alert(msg, getLocation())
        relLogSem()
    except Exception, e:
        relLogSem()
        raise e   


def setDebug(debug):
    """
    Set Global Debug Flag.

    debug:    Debug flag. 1 = debug mode (integer).

    Returns:  Void.
    """
    global _debug
    _debug = debug


def getDebug():
    """
    Return Global Debug Flag.
    
    Returns:  Debug flag. 1 = debug mode (integer).
    """
    global _debug
    return _debug


def DEBUG(msg = ""):
    """
    Function to use for debugging purposes. Print a message on stdout.

    msg:       Message to print (string).

    Returns:   Void.
    """
    if (getDebug()):
        import traceback
        stackInfo = traceback.extract_stack()[-2]
        module = stackInfo[0].split("/")[-1]
        lineNo = stackInfo[1]
        method = stackInfo[2]
        print "##### DEBUG: Module: " + module + " - Line No: " +\
              str(lineNo) + " - Method: " + method + " - Message: " + str(msg)


def logFlush():
    """
    Flush the logs cached in the log manager.

    Returns:    Void.
    """
    PccLog.__logMgr.flush()


def setLogCache(size):
    """
    Set the log cache to a specific size.

    size:     Number of log entries to cache before flushing (int).

    Returns:  Void.
    """
    PccLog.__logMgr.setLogCacheFlushSize(size)


def getAttribValue(node,
                   attributeName,
                   ignoreFailure = 0):
    """
    Get value of an attribute in connection with a DOM Node.

    node:            XML node object (Node).

    attributeName:   Name of attribute (string).

    ignoreFailure:   If a failure (execption) occurrs the function
                     will not raise an exception but will return with
                     '' as return value (integer

    Returns:         Value of attribute (string).
    """
    if (not ignoreFailure):
        try:
            return str(node._attrs[attributeName].nodeValue)
        except Exception, e:
            errMsg = "Error retrieving value for attribute: " + attributeName+\
                     " from node: " + node.nodeName + ". Error: " + str(e)
            raise Exception, errMsg
    else:
        try:
            return str(node._attrs[attributeName].nodeValue)
        except:
            return ""


def ngamsGetChildNodes(parentNode,
                       childNodeName):
    """
    Get the child nodes of the parent node DOM object given, which have
    the given 'childNodeName'.

    parentNode:     Parent node DOM object (Node).

    childNodeName:  Name of chile node (element) (string).

    Returns:        List with chile node DOM objects (list/Node).
    """
    childNodes = []
    for childNode in parentNode.childNodes:
        try:
            nodeName = childNode.nodeName
        except:
            nodeName = ""
        if (nodeName == childNodeName):
            childNodes.append(childNode)
    return childNodes


def getHostName():
    """
    Return the host name is it should be used internally in NG/AMS.

    Returns:   Host name for this NGAS System (string).
    """
    hostName = os.uname()[1]
    if (hostName.split(".")[-1] == "local"):
        return hostName.split(".")[0].split("-")[0]
    else:
        return hostName.split(".")[0]


def setSrvPort(portNo):
    """
    Set the server port number. Should only be set if multiple NG/AMS Servers
    are executed on the same node.

    portNo:   Server port number (integer).

    Returns:  Void.
    """
    global _srvPortNo
    _srvPortNo = int(portNo)


def getSrvPort():
    """
    Get the server port number. Is only set if multiple NG/AMS Servers
    are executed on the same node.

    Returns:  Server port number (integer|None).
    """
    global _srvPortNo
    return _srvPortNo


def getHostId():
    """
    Returns the proper NG/AMS Host ID according whether multiple servers
    can be executed on the same host.

    If multiple servers can be executed on one node, the Host ID will be:

      <Host Name>:<Port No>

    Otherwise, the Host ID will simply be the host name.

    Returns:    NG/AMS Host ID (string).
    """
    if (getSrvPort()):
        return getHostName() + ":" + str(_srvPortNo)
    else:
        return getHostName()


def ignoreValue(ignoreEmptyField,
                fieldValue):
    """
    Evaluate the value of a field and indicate whether to ignore it or not.

    Internal function.
    
    ignoreEmptyField:     If set to 1 indicates that the field should be 
                          ignored if it is an empty string (integer/0|1).
                          
    fieldValue:           Value of field (string). 

    Returns:              1 if field should be ignored, otherwise 0
                          (integer/0|1).
    """
    if (type(fieldValue) == types.IntType):
        if (ignoreEmptyField and (fieldValue == -1)): return 1
    elif (ignoreEmptyField and (str(fieldValue).strip() == "")):
        return 1
    return 0


def getFileSize(filename):
    """
    Get size of file referred.

    filename:   Filename - complete path (string).

    Returns:    File size (integer).
    """
    #return int(os.stat(filename)[6])
    return os.path.getsize(filename)


def getFileAccessTime(filename):
    """
    Get last access time of file referred.

    filename:   Filename - complete path (string).

    Returns:    Last access time (seconds since epoch) (integer).
    """
    #IMPL: Use this: return os.path.getatime(filename)
    return int(os.stat(filename)[7])


def getFileCreationTime(filename):
    """
    Get creation time of file referred.

    filename:   Filename - complete path (string).

    Returns:    Creation time (seconds since epoch) (integer).
    """
    # IMPL: Use this: return os.path.getctime(filename)
    return int(os.stat(filename)[9])


def getFileModificationTime(filename):
    """
    Get the last modification date of the file (seconds since epch).

    filename:   Filename - complete path (string).

    Returns:    File modification date (integer).
    """
    # IMPL: Use this: return os.path.getmtime(filename)
    return int(os.stat(filename)[8])


def getUniqueNo():
    """
    Generate a unique number (unique for this session of NG/AMS).

    Returns:  Unique number (integer).
    """
    global _uniqueNumberSem, _uniqueNumberCount
    _uniqueNumberSem.acquire()
    _uniqueNumberCount += 1
    count = _uniqueNumberCount
    _uniqueNumberSem.release()
    return count


def cleanList(lst):
    """
    Remove empty elements from a list containing strings
    (elements of length 0).

    lst:      List to be cleaned (list).

    Returns:  Cleaned list (list).
    """
    cleanLst = []
    for el in lst:
        if (len(el) > 0): cleanLst.append(el)
    return cleanLst


def createSortDicDump(dic):
    """
    Create a sorted ASCII representation of a dictionary. The order is sorted
    according to the dictionary keys.

    dic:     Source dictionary (dictionary).

    Returns: Sorted dictionary ASCII representation (string).
    """
    if (type(dic) != types.DictType):
        raise Exception, "Object given is not a dictionary"
    keys = dic.keys()
    keys.sort()
    asciiDic = ""
    for key in keys: asciiDic += ", '%s': '%s'" % (key, dic[key])
    asciiDic = "{" + asciiDic[2:] + "}"
    return asciiDic


def getDiskSpaceAvail(mountPoint,
                      format = "MB",
                      float = 0):
    """
    Get the disk space available for the disk with the given mount point.

    mountPoint:  Mount point (string).

    format:      Output format (B|KB|MB|GB|TB/string).

    float:       Return the result in floating point (0|1/integer).

    Returns:     Returns available space in MB (integer).
    """
    info(4,"Checking disk space available for path: " + mountPoint + " ...")
    if (not os.path.exists(mountPoint)): os.makedirs(mountPoint)
    if (not os.path.exists(mountPoint)):
        raise Exception, "Illegal path encountered: " +\
              mountPoint + " - or path could not be created."
    # IMPL: UNIX specific/make plug-in?
    exitCode, stdOut = commands.getstatusoutput("df -k %s " % mountPoint)
    if (exitCode):
        raise Exception, "Illegal path encountered: " +\
              mountPoint + " - or path could not be created."
    diskSpace = eval(str(cleanList(stdOut.replace("\n", " ").\
                                   split(" "))[10]) + ".0")
    scale = {"B": 1024., "KB": 1., "MB": 1./1024.,
             "GB": 1./1048576., "TB": 1./1073741824.}
    diskSpace = (diskSpace * scale[format])

    if (not float): diskSpace = int(diskSpace + 0.5)
    msg = "Checked disk space available for path: %s - Result (%s): %s"
    info(4,msg % (mountPoint, format, str(diskSpace)))
    return diskSpace


def checkAvailDiskSpace(filename,
                        fileSize):
    """
    Check if there is enough space to store the file. If there is not
    enough space, an exception is raised.
    
    filename:   Filename - complete path (string).
    
    fileSize:   Filesize in bytes (integer).

    Returns:    Void.
    """
    path = os.path.dirname(filename)
    info(4,"Checking for disk space availability for path: " + path +\
         " - Needed size: " + str(fileSize) + " ...")    
    if ((fileSize / 1024**2) > getDiskSpaceAvail(path)):
        errMsg = genLog("NGAMS_ER_NO_DISK_SPACE", [filename, fileSize])
        error(errMsg)
        raise Exception, errMsg


def checkCreatePath(path):
    """
    Check if the path referred exists - if not it is created.

    path:      Path to check/create (string).

    Returns:   Void.
    """
    # Check if path exists, if not, create it.
    if (os.path.exists(path) == 0):
        info(4,"Creating non-existing path: " + path)
        try:
            os.makedirs(path)
            os.chmod(path, 0775)
        except Exception, e:
            error("Error creating path: " + str(e))
            raise e

    
def rmFile(filename):
    """
    Remove the file referenced.

    filename:   File to remove (string).

    Returns:    Void.
    """
    info(4,"Removing file: %s" % filename)
    commands.getstatusoutput("rm -rf " + filename)


def mvFile(srcFilename,
           trgFilename):
    """
    Move a file from the source filename to the specified target filename.

    srcFilename:  Source filename (string).
    
    trgFilename:  Target filename (string).
    
    Returns:      Time in took to move file (s) (float).
    """
    info(4,"Moving file: " + srcFilename + " to filename: " + trgFilename)
    try:
        # Make target file writable if existing.
        if (os.path.exists(trgFilename)): os.chmod(trgFilename, 420)
        checkCreatePath(os.path.dirname(trgFilename))
        fileSize = getFileSize(srcFilename)
        checkAvailDiskSpace(trgFilename, fileSize)
        timer = PccUtTime.Timer()
        stat, out = commands.getstatusoutput("mv %s %s" %\
                                             (srcFilename, trgFilename))
        if (stat): raise Exception, "Error executing move command: " + str(out)
        deltaTime = timer.stop()
    except Exception, e:
        errMsg = genLog("NGAMS_AL_MV_FILE", [srcFilename, trgFilename, str(e)])
        alert(errMsg)
        raise Exception, errMsg
    info(4,"File: " + srcFilename + " moved to filename: " + trgFilename)

    return deltaTime


def cpFile(srcFilename,
           trgFilename):
    """
    Copy a file from the source filename to the specified target filename.

    srcFilename:  Source filename (string).
    
    trgFilename:  Target filename (string).
    
    Returns:      Time in took to move file (s) (float).
    """
    info(4,"Copying file: " + srcFilename + " to filename: " + trgFilename)
    try:
        # Make target file writable if existing.
        if (os.path.exists(trgFilename)): os.chmod(trgFilename, 420)
        checkCreatePath(os.path.dirname(trgFilename))
        fileSize = getFileSize(srcFilename)
        checkAvailDiskSpace(trgFilename, fileSize)
        timer = PccUtTime.Timer()
        stat, out = commands.getstatusoutput("cp %s %s" %\
                                             (srcFilename, trgFilename))
        if (stat): raise Exception, "Error executing copy command: " + str(out)
        deltaTime = timer.stop()
    except Exception, e:
        errMsg = genLog("NGAMS_AL_CP_FILE", [srcFilename, trgFilename, str(e)])
        alert(errMsg)
        raise Exception, errMsg
    info(4,"File: %s copied to filename: %s" % (srcFilename, trgFilename))
    return deltaTime
    

def timeRef2Iso8601(timeRef):
    """
    Convert a time reference into ISO8601. This can be given as

      o ISO8601.
      o Seconds since epoch.
      o DateTime time stamp.

    If None or an empty string is given, an empty string is returned.

    timeRef:    The reference to the time (string|int|DateTime).

    Returns:    ISO8601 timestamp (string).
    """
    if (not timeRef):
        timeRefConv = ""
    elif (str(timeRef).find(":") != -1):
        timeRef = str(timeRef)
        # Assume ISO8601 format or "YYYY-MM-DD HH:MM:SS[.sss]".
        if (timeRef.find("T") == -1):
            idx = timeRef.find(" ")
            timeRefConv = timeRef[0:idx] + "T" + timeRef[(idx + 1):]
        else:
            timeRefConv = timeRef
    elif (type(timeRef) == types.IntType):
        # Assume seconds from epoch.
        timeRefConv = PccUtTime.TimeStamp().initFromSecsSinceEpoch(timeRef).\
                      getTimeStamp()
    else:
        # Assume DateTime object.
        timeRefConv = PccUtTime.TimeStamp().\
                      initFromSecsSinceEpoch(timeRef.ticks()).\
                      getTimeStamp()
    return timeRefConv


def padString(strBuf,
              reqLen,
              prependChr):
    """
    Prepend a certain character to generate a string of a certain length.
    E.g. if strBuf='12', reqLen=6 and prependChr='0', the resulting string
    will be '000012'.

    strBuf:         String buffer to check/change (string).
    
    reqLen:         Desired lenght of string (integer).
    
    prependChr:     Character to prepend (string/length=1).

    Returns:        Resulting string with the characters prepended (string).
    """
    noMisChars = (reqLen - len(strBuf))
    for i in range(noMisChars):
        strBuf = prependChr + strBuf
    return strBuf


def loadDoc(docName):
    """
    Load the given documation page and return the contents in a string buffer.
    Lines initiated with # in the documentation are filtered out.

    docName:  Name of documentation page, relative to the ngams module
              root directory (string).

    Returns:  Documentation page (string).
    """
    complDocName = os.path.normpath(ngamsGetSrcDir() + "/" + docName)
    fo = open(complDocName)
    docBufLines = fo.readlines()
    fo.close()
    docBuf = ""
    for docLine in docBufLines:
        if (docLine[0] != "#"): docBuf += docLine
    return docBuf


def setTestMode():
    """
    Switch on the Test Mode Flag. When set, special conditions are applied
    during execution of SW/server.
    
    Returns:   Void.
    """
    global _testMode
    _testMode = 1
        
    
def getTestMode():
    """
    Return the value of the Test Mode Flag. When set to 1, the server will not
    kill itself (as this would kill the test case).

    Returns:     Test Mode Flag (integer/0|1).
    """
    global _testMode
    return _testMode


# EOF
