#******************************************************************************
# ESO/DFS
#
# "@(#) $Id: ngamsDataCheckThread.py,v 1.74 2005/08/30 11:18:29 jknudstr Exp $"
#
# Who       When        What
# --------  ----------  -------------------------------------------------------
# jknudstr  29/01/2002  Created
#
    

"""
This module contains the code for the Data Check Thread, which is used
to check the data holding in connection with one NGAS host.
"""

import os, time, thread, threading, random, glob
import pcc, PccUtTime
from ngams import *
import ngamsDb, ngamsDbm, ngamsLib, ngamsHighLevelLib, ngamsNotification
import ngamsDiskInfo, ngamsFileUtils


def startDataCheckThread(srvObj):
    """
    Start the Data Check Thread.

    srvObj:     Reference to server object (ngamsServer).
    
    Returns:    Void.
    """
    if (not srvObj.getCfg().getDataCheckActive()): return
    info(3,"Starting Data Check Thread ...")
    srvObj._dataCheckRunSync.set()
    args = (srvObj, None)
    srvObj._dataCheckThread = threading.Thread(None, dataCheckThread,
                                               NGAMS_DATA_CHECK_THR, args)
    srvObj._dataCheckThread.setDaemon(0)
    srvObj._dataCheckThread.start()
    info(3,"Data Check Thread started")
 

def stopDataCheckThread(srvObj):
    """
    Stop the Data Check Thread.

    srvObj:     Reference to server object (ngamsServer).
    
    Returns:    Void.
    """
    if (not srvObj.getCfg().getDataCheckActive()): return
    info(3,"Stopping Data Check Thread ...")
    srvObj._dataCheckRunSync.clear()
    srvObj._dataCheckRunSync.wait(10)
    srvObj._dataCheckRunSync.clear()
    srvObj._dataCheckThread = None
    info(3,"Data Check Thread stopped")
  

def _stopDataCheckThr(srvObj):
    """
    Convenience function used by the DCC Thread to check if it is allowed
    to run and in case not, to stop itself.

    srvObj:      Reference to instance of ngamsServer object (ngamsServer).

    Returns:     Void.
    """
    if (not srvObj._dataCheckRunSync.isSet()):
        info(2,"Stopping Data Check Thread")
        srvObj.updateHostInfo(None, None, None, None, None, None, 0, None)
        srvObj._dataCheckRunSync.set()
        raise Exception, "_STOP_DATA_CHECK_THREAD_"
  
 
# IMPL: Variables used for the handling of the execution of the DCC within
# this module. Could be made members of the ngamsServer class, but since they
# are only used locally, they are defined as global within this module
# in order not to complicate the ngamsServer class.
# Alternatively, one global dictionary could be used with keys:
# "_diskDic" -> "DiskDic", "_dbmObjDic" -> "DbmObjDic", ...
_diskDic         = {}    # Dictionary with info about disks concerned.
_dbmObjDic       = {}    # Dictionary with Queue/Error DBM objects.
_diskIdList      = []    # List of Disk IDs ordered for the checking.
_diskSchedDic    = {}    # Dictionary used when scheduling disks to be checked.
_fileRefDbm      = {}    # Points to DBM with references to all files on disks.
_seqFileCheck    = 1     # If one file should be checked sequentially.
_reqFileInfoSem  = threading.Semaphore(1)  # Semaphore for critical access.
_runSubThread    = 1     # Flag to signal to sub-threads to run or not.

# Parameters for statistics.
_statCheckSem     = threading.Semaphore(1)  # Semaphore to protect checking.
_statLastDbUpdate = 0        # Time (secs since epoch) for last DB update.
_statCheckStart   = None     # Start time for checking (secs since epoch).
_statCheckRemain  = None     # Remaining time in seconds.
_statCheckRate    = None     # Speed of the data checking (MB/s).
_statCheckMb      = None     # Amount of data to check (MB).
_statCheckedMb    = None     # Amount of data checked (MB).
_statCheckFiles   = None     # Number of files to check.
_statCheckCount   = None     # Number of files checked.


def _getDiskDic():
    """
    Return refrence to the Disk Dictionary with information about the disks
    concerned.
    
    Returns:    Disk Dictionary (dictionary).
    """
    global _diskDic
    return _diskDic


def _resetDiskDic():
    """
    Reset the Disk Dictionary.

    Returns:   Void.
    """
    global _diskDic
    _diskDic = {}


def _getDbmObjDic():
    """
    Return reference to Queue/Error DBM Objects Dictionary.

    Returns:    Reference to DBM Objects Dictionary (dictionary).
    """
    global _dbmObjDic
    return _dbmObjDic


def _resetDbmObjDic():
    """
    Reset the DBM Object Dictionary.

    Returns:   Void.
    """
    global _dbmObjDic
    _dbmObjDic = {}
    

def _getDiskIdList():
    """
    Return reference to Queue/Error DBM Disk ID List.

    Returns:    Reference to DBM Slot List (list).
    """
    global _diskIdList
    return _diskIdList


def _resetDiskIdList():
    """
    Reset the Disk ID List.

    Returns:    Void.
    """
    global _diskIdList
    _diskIdList = []


def _getDiskSchedDic():
    """
    Return reference to the Disk Scheduling Dictionary.

    Returns:    Reference to DBM Slot List (list).
    """
    global _diskSchedDic
    return _diskSchedDic


def _resetDiskSchedDic():
    """
    Reset the Disk Scheduling Dictionary.

    Returns:    Void.
    """
    global _diskSchedDic
    _diskSchedDic = {}


def _setFileRefDbm(dbmObj):
    """
    Set the File Reference Object.

    dbmObj:    DBM object (ngamsDbm).

    Returns:   Void.
    """
    global _fileRefDbm
    _fileRefDbm = dbmObj


def _getFileRefDbm():
    """
    Return reference to DBM containing references to all files on the
    storage disks.

    Returns:   Reference to DBM object (ngamsDbm).
    """
    global _fileRefDbm
    return _fileRefDbm


def _resetFileRefDbm():
    """
    Reset the File Reference DBM.

    Returns:   Void.
    """
    global _fileRefDbm
    if (_fileRefDbm): del _fileRefDbm
    _fileRefDbm = {}
    

def _getSeqFileCheck():
    """
    Return the flag indicating if the files should be checked randomly or
    sequentially.

    Returns:   1 = sequential checking, 0 = random checking (integer/0|1).
    """
    global _seqFileCheck
    return _seqFileCheck


def _setSeqFileCheck(seqFileCheck):
    """
    Set the flag indicating if files should be chaked sequentially.

    seqFileCheck:  1 = sequential checking, 0 = random checking (integer/0|1).

    Returns:       Void.
    """
    global _seqFileCheck
    _seqFileCheck = seqFileCheck


def _takeReqFileInfoSem():
    """
    Take the semphore protecting the access to the file info DBMs.

    Returns:    Void.
    """
    global _reqFileInfoSem
    _reqFileInfoSem.acquire()


def _relReqFileInfoSem():
    """
    Release the semphore protecting the access to the file info DBMs.

    Returns:    Void.
    """
    global _reqFileInfoSem
    _reqFileInfoSem.release()


def _initFileCheckStatus(srvObj,
                         amountMb,
                         noOfFiles):
    """
    Initialize the checking parameters.

    srvObj:         Reference to instance of ngamsServer object (ngamsServer).

    amountMb:       Amount of data to check (MB) (float).
    
    noOfFiles:      Number of files to check (integer).

    Returns:        Void.
    """
    info(4,"Entering _initFileCheckStatus() ...")
    global _statCheckSem, _statLastDbUpdate, _statCheckStart,\
           _statCheckRemain, _statCheckRate, _statCheckMb, _statCheckedMb,\
           _statCheckFiles, _statCheckCount
    try:
        _statCheckSem.acquire()
        _statCheckStart   = time.time()
        _statCheckRemain  = 0
        statEstimTime     = 0
        _statCheckRate    = 0.0
        _statCheckMb      = amountMb
        _statCheckedMb    = 0.0
        _statCheckFiles   = noOfFiles
        _statCheckCount   = 0
        _statLastDbUpdate = 0 
        _statCheckSem.release()
    except Exception, e:
        _statCheckSem.release()
        error("Exception caught in Data Checking Thread: " + str(e))
    srvObj.getDb().updateDataCheckStat(getHostId(), _statCheckStart,
                                       _statCheckRemain, statEstimTime,
                                       _statCheckRate, _statCheckMb,
                                       _statCheckedMb, _statCheckFiles,
                                       _statCheckCount)
    info(4,"Leaving _initFileCheckStatus()")


def _updateFileCheckStatus(srvObj,
                           fileSize,
                           diskId,
                           fileId,
                           fileVersion,
                           report,
                           force = 0):
    """
    Update the status of the DCC.

    srvObj:       Reference to instance of ngamsServer object (ngamsServer).

    fileSize:     Size of file (in bytes) that was checked (integer)

    diskId:       ID of disk hosting file checked (string).
    
    fileId:       ID of file concerned (string).
    
    fileVersion:  Version of file concered (integer).
    
    report:       List containing the result of the file checking.
                  Refer to documentation for ngamsFileUtils.checkFile()
                  for futher information (list).

    force:        If set to 1 a DB update will be forced (integer/0|1).

    Returns:      Void.
    """
    info(5,"Entering _updateFileCheckStatus() ...")
    global _statCheckSem, _statLastDbUpdate, _statCheckStart,\
           _statCheckRemain, _statCheckRate, _statCheckMb, _statCheckedMb,\
           _statCheckFiles, _statCheckCount
    try:
        _statCheckSem.acquire()
        timeNow = time.time()

        # Calculate the new values.
        if (fileId):
            _statCheckedMb  += (float(fileSize) / 1048576.0)
            _statCheckCount += 1

        checkTime        = (timeNow - _statCheckStart)
        _statCheckRate   = (float(_statCheckedMb) / float(checkTime))
        if ( _statCheckRate > 0):
            _statCheckRemain = int((float(_statCheckMb - _statCheckedMb) /
                                    _statCheckRate) + 0.5)
            statEstimTime    = int(float(_statCheckMb / _statCheckRate) + 0.5)
        else:
            _statCheckRemain = 0
            statEstimTime    = 0

        # Update DB only every 10s.
        if (force or ((timeNow - _statLastDbUpdate) >= 10)):
            srvObj.getDb().updateDataCheckStat(getHostId(), _statCheckStart,
                                               _statCheckRemain, statEstimTime,
                                               _statCheckRate, _statCheckMb,
                                               _statCheckedMb, _statCheckFiles,
                                               _statCheckCount)
            _statLastDbUpdate = timeNow

        # Update report if an error was found.
        if (diskId and report):
            fileKey = ngamsLib.genFileKey(None, fileId, fileVersion)
            _getDbmObjDic()[diskId][1].add(fileKey, report)
            
        statFormat = "DCC Status: Time Remaining (s): %d, " +\
                     "Rate (MB/s): %.3f, " +\
                     "Volume/Checked (MB): %.3f/%.3f, Files/Checked: %d/%d"
        info(4,statFormat % (_statCheckRemain, _statCheckRate, _statCheckMb, 
             _statCheckedMb, _statCheckFiles, _statCheckCount))
        _statCheckSem.release()
    except Exception, e:
        error("Data Consistency Checking: Encountered error: " + str(e))
        _statCheckSem.release()
    info(5,"Leaving _updateFileCheckStatus()")
    

def _setRunPermSubThread(perm):
    """
    Set the run permission for the DCC Sub-Thread.

    perm:     Permission: 1 = run, 0 = stop (integer/0|1).

    Returns:  Void.
    """
    global _runSubThread
    _runSubThread = perm


def _getRunPermSubThread():
    """
    Get the run permission for the DCC Sub-Thread.

    Returns:     Permission to tun: 1 = run, 0 = stop (integer/0|1).
    """
    global _runSubThread
    return _runSubThread

    
def _stopCheckingSubThread():
    """
    Used by the Data Consistency Checking Sub-Threads to check if they are
    allowed to execute.

    Returns:   Void.
    """
    if (not _getRunPermSubThread()):
        info(3,"DCC Sub-Thread exiting ...")
        thread.exit()


def _suspend(srvObj,
             baseTime = 0.010):
    """
    Determine the time the thread should suspend itself according to
    the load, priority etc. + suspend for the given time.

    srvObj:      Reference to instance of ngamsServer object (ngamsServer).

    baseTime:    Time used to calculate the suspension time (float).

    Returns:     Void.
    """
    suspTime = (srvObj.getCfg().getDataCheckPrio() * baseTime)
    time.sleep(suspTime)
    

def _dumpFileInfo(srvObj,
                  tmpFilePat):
    """
    Function that dumps the information about the files. One DBM is created
    per disk. This is named:

       <Mount Root Point>/cache/DATA-CHECK-THREAD_QUEUE_<Disk ID>.bsddb

    If problems are found for a file, these are stored in DBM files named:

       <Mount Root Point>/cache/DATA-CHECK-THREAD_ERRORS_<Disk ID>.bsddb

    A DBM is also created with references to all files found on the system.
    The name of this is of the form:

       <Mount Root Point>/cache/DATA-CHECK-THREAD_FILES_<Host ID>.bsddb

    Latter is always generated from scratch.


    The function handles the DBM files in the following way:

       1. Get the information about the disks in this system.
       
       2. Check for each DBM file found, if this disk is still in the system.
          If not, the Queue and Error DBM files are removed.

       3. Go through the list of disks in the system. If they don't have
          the two DBM files listed about, these are initialized. The file
          information for all the files stored on the disk is dumped into
          the Queue DBM file. Only files marked to be ignored are not dumped.

       4. Finally, build up a DBM with references to all files found
          on this system
 
    srvObj:       Reference to server object (ngamsServer).

    tmpFilePat:   Pattern for temporary files (string).
     
    Returns:      Void.
    """
    info(4,"Entering _dumpFileInfo() ...")

    cacheDir = srvObj.getCfg().getMountRootDirectory()+ "/" + NGAMS_CACHE_DIR
    checkCreatePath(os.path.normpath(cacheDir))

    ###########################################################################
    # Get information about all disks in the system. The disks are only
    # added if 'ripe' for checking.
    ###########################################################################
    info(4,"Get information about all disks mounted in this system ...")
    slotIdList = srvObj.getDb().getSlotIdsMountedDisks(getHostId())
    diskListRaw = srvObj.getDb().\
                  getDiskInfoForSlotsAndHost(getHostId(), slotIdList)
    minCycleTime=ngamsLib.isoTime2Secs(srvObj.getCfg().getDataCheckMinCycle())
    _resetDiskDic()
    lastDiskCheckDic = {}
    for diskInfo in diskListRaw:
        _stopDataCheckThr(srvObj)
        tmpDiskInfoObj = ngamsDiskInfo.ngamsDiskInfo().\
                         unpackSqlResult(diskInfo)
        if (not tmpDiskInfoObj.getLastCheck()):
            lastCheck = 0
        else:
            lastCheck = ngamsLib.iso8601ToSecs(tmpDiskInfoObj.getLastCheck())
        if ((time.time() - lastCheck) >= minCycleTime):
            if (lastCheck):
                if (not lastDiskCheckDic.has_key(lastCheck)):
                    lastDiskCheckDic[lastCheck] = tmpDiskInfoObj
                else:
                    lastDiskCheckDic[lastCheck + len(lastDiskCheckDic)] =\
                                               tmpDiskInfoObj
            else:
                lastDiskCheckDic[time.time()] = tmpDiskInfoObj
            _getDiskDic()[tmpDiskInfoObj.getDiskId()] = tmpDiskInfoObj
    info(4,"Got information about all disks mounted in this system")
    ###########################################################################
            
    ###########################################################################
    # Loop over the Queue/Error DBM files found, check if the disk is
    # still in the system/scheduled for checking.
    ###########################################################################
    info(4,"Loop over/check existing Queue/Error DBM Files ...")
    dbmFileList = glob.glob(cacheDir + "/" + NGAMS_DATA_CHECK_THR +\
                            "_QUEUE_*.bsddb")
    for dbmFile in dbmFileList:
        _stopDataCheckThr(srvObj)
        diskId = dbmFile.split("_")[-1].split(".")[0]
        if (not _getDiskDic().has_key(diskId)):
            filePat = "%s/%s*%s.bsddb" % (cacheDir,NGAMS_DATA_CHECK_THR,diskId)
            rmFile(filePat)
        else:
            # Add references to Queue/Error DBM.
            queueDbmFile = "%s/%s_QUEUE_%s.bsddb" %\
                           (cacheDir, NGAMS_DATA_CHECK_THR, diskId)
            queueDbm = ngamsDbm.ngamsDbm(queueDbmFile, 0, 1)
            errorDbmFile = "%s/%s_ERRORS_%s.bsddb" %\
                           (cacheDir, NGAMS_DATA_CHECK_THR, diskId)
            errorDbm = ngamsDbm.ngamsDbm(errorDbmFile, 0, 1)
            _getDbmObjDic()[diskId] = (queueDbm, errorDbm)
    info(4,"Looped over/checked existing Queue/Error DBM Files")
    ###########################################################################

    ###########################################################################
    # Loop over the disks mounted in this system and check if they have a
    # Queue/Error DBM file. In case the DBM files are not available, create
    # these.
    ###########################################################################
    info(4,"Create DBM files for disks to be checked ...")
    startDbFileRd = time.time()
    for diskId in _getDiskDic().keys():
        _stopDataCheckThr(srvObj)
        if (_getDbmObjDic().has_key(diskId)): continue
        
        # The disk is ripe for checking but still has no Queue/Error DBM
        # DBs allocated.
        queueDbmFile = "%s/%s_QUEUE_%s.bsddb" %\
                       (cacheDir, NGAMS_DATA_CHECK_THR, diskId)
        tmpQueueDbmFile = tmpFilePat + "_" + os.path.basename(queueDbmFile)
        queueDbm = ngamsDbm.ngamsDbm(tmpQueueDbmFile, 0, 1)

        # Get the theoretical number of files.
        noOfFilesList = []
        accuNoOfFiles = 0
        for n in range(3):
            noOfFilesList.append(srvObj.getDb().getNumberOfFiles(diskId,
                                                                 ignore=0))
            accuNoOfFiles += int(noOfFilesList[0])
            time.sleep(0.1)
        if ((3 * int(noOfFilesList[0])) != accuNoOfFiles):
            errMsg = "Problem querying number of files: %d/%d/%d" %\
                     (noOfFilesList[0], noOfFilesList[1], noOfFilesList[2])
            errMsg = genLog("NGAMS_ER_DB_COM", [errMsg])
            raise Exception, errMsg
        expNoOfFiles = noOfFilesList[0]
        
        # Now, retrieve the files on the given disk, and store the info
        # in the Queue DBM file.
        actNoOfFiles = 0
        # IMPL: Use ngamsDb.dumpFileSummary1().
        cursorObj = srvObj.getDb().getFileSummary1(None, [diskId], [],
                                                   ignore=0, fileStatus=[],
                                                   lowLimIngestDate=None)
        if (getTestMode()):
            # Would like to obtain ~10% chance/risk of failure during dumping
            # of file info.
            fetchSize = ((expNoOfFiles / 2) + 0.5)
            if (fetchSize > 1000): fetchSize = 1000
        else:
            fetchSize = 1000
        while (1):
            _stopDataCheckThr(srvObj)
            try:
                fileList = cursorObj.fetch(fetchSize)
            except Exception, e:
                # Assume a problem like e.g. a broken DB connection.
                errMsg = "Problem encountered while dumping file info. " +\
                         "Error: " + str(e)
                error(errMsg)
                del queueDbm
                rmFile(tmpQueueDbmFile)
                raise Exception, genLog("NGAMS_ER_DB_COM", [errMsg])
                
            if (not fileList): break
            actNoOfFiles += len(fileList)
            for fileInfo in fileList:
                fileId  = fileInfo[ngamsDb.SUM1_FILE_ID]
                fileVer = fileInfo[ngamsDb.SUM1_VERSION]
                fileKey = ngamsLib.genFileKey(None, fileId, fileVer)
                queueDbm.add(fileKey, fileInfo)
            queueDbm.sync()
            _suspend(srvObj)
        queueDbm.sync()
        del cursorObj

        # Check if the expected number of files is equal to the actual
        # number of files dumped into the DBM.
        if (actNoOfFiles != expNoOfFiles):
            errMsg = "Number of files dumped for disk: %s: %d, differs " +\
                     "from expected number: %d"
            errMsg = errMsg % (diskId, actNoOfFiles, expNoOfFiles)
            error(errMsg)
            rmFile(tmpQueueDbmFile)
            del queueDbm
            raise Exception, genLog("NGAMS_ER_DB_COM", [errMsg])

        # Rename DCC Queue DBM from the temporary to the final name.
        del queueDbm
        mvFile(tmpQueueDbmFile, queueDbmFile)
        queueDbm = ngamsDbm.ngamsDbm(queueDbmFile, 0, 1)

        # Create Error DBM + add these in the DBM Dictionary for the disk.
        errorDbmFile = "%s/%s_ERRORS_%s.bsddb" %\
                       (cacheDir, NGAMS_DATA_CHECK_THR, diskId)
        errorDbm = ngamsDbm.ngamsDbm(errorDbmFile, 0, 1)
        _getDbmObjDic()[diskId] = (queueDbm, errorDbm)
        
        _stopDataCheckThr(srvObj)
    info(3,"Queried info for files to be checked from DB. Time: %.3fs" %\
         (time.time() - startDbFileRd))
    info(4,"Checked that disks scheduled for checking have DBM files")
    ###########################################################################

    ###########################################################################
    # Create a DBM with references to all files found on this system. The
    # complete path name are the keys in the dictionary. While going through
    # the files registered in the DB, we remove the entries in this dictionary.
    # The remaining entries in the dictionary, are thus files, which are not
    # registered in the DB.
    #
    # We try with "<mt pt>/*", "<mt pt>/*/*", "<mt pt>/*/*/*", ...,
    # - until no files are found.
    ###########################################################################
    info(4,"Create DBMs with references to all files on the storage disks ...")
    fileRefDbm = "%s/%s_FILES_%s.bsddb" %\
                 (cacheDir, NGAMS_DATA_CHECK_THR, getHostId())
    rmFile(fileRefDbm)
    _setFileRefDbm(ngamsDbm.ngamsDbm(fileRefDbm, 1, 1))
    tmpGlobFile = tmpFilePat + "_FILE_GLOB.glob"
    for diskId in _getDiskDic().keys():
        _stopDataCheckThr(srvObj)
        diskInfoObj = _getDiskDic()[diskId]
        pattern = ""
        while (1):
            # Use a shell command here to avoid building up maybe huge
            # lists in memory/in the Python interpreter. A better way could
            # maybe be found avoiding to invoke a shell command.
            rmFile(tmpGlobFile)
            searchPath = os.path.normpath(diskInfoObj.getMountPoint()+pattern)
            tmpCmd = "find " + searchPath + " -maxdepth 1 > " + tmpGlobFile
            stat, out = commands.getstatusoutput(tmpCmd)
            if (stat != 0): break
            fo = open(tmpGlobFile)
            while (1):
                _stopDataCheckThr(srvObj)
                nextFile = fo.readline().strip()
                if (not nextFile): break
                # NGAS Disk Info file + files under the staging area on
                # the disks are ignored.
                if (os.path.isfile(nextFile) and
                    (os.path.basename(nextFile) != NGAMS_DISK_INFO) and
                    (os.path.basename(nextFile) != NGAMS_VOLUME_ID_FILE) and
                    (nextFile.find("/" + NGAMS_STAGING_DIR + "/") == -1)):
                    _getFileRefDbm().add(nextFile, [diskInfoObj.getDiskId()])
            fo.close()
            pattern += "/*"
        _getFileRefDbm().sync()
        _stopDataCheckThr(srvObj)
        rmFile(tmpGlobFile)
    info(4,"Created DBMs with references to all files on the storage disks")
    ###########################################################################

    ###########################################################################
    # Now go through files to be ignored. These are also removed if found in
    # the File Reference DBM. This is done, since these files are not
    # taking into account in the checking loop
    ###########################################################################
    info(4,"Retrieve information about files to be ignored ...")
    spuFilesCur = srvObj.getDb().getFileSummarySpuriousFiles1(getHostId())
    while (1):
        _stopDataCheckThr(srvObj)
        fileList = spuFilesCur.fetch(1000)
        if (not fileList): break

        # Loop over the files.
        for fileInfo in fileList:
            if (fileInfo[ngamsDb.SUM1_FILE_IGNORE]):
                filename = os.path.\
                           normpath(fileInfo[ngamsDb.SUM1_MT_PT] + "/" +\
                                    fileInfo[ngamsDb.SUM1_FILENAME])
                if (_getFileRefDbm().hasKey(filename)):
                    _getFileRefDbm().rem(filename)
        _suspend(srvObj)
    _getFileRefDbm().sync()
    del spuFilesCur
    info(4,"Retrieved information about files to be ignored")
    ###########################################################################

    ###########################################################################
    # Order the Disk IDs of the disks to be checked according to the
    # checking sequence scheme defined.
    ###########################################################################
    info(4,"Order disks according to the sequence they should be checked ...")
    if (srvObj.getCfg().getDataCheckDiskSeq() == NGAMS_CHECK_SEQ):
        lastCheckList = lastDiskCheckDic.keys()
        lastCheckList.sort()
        _resetDiskIdList()
        for lastCheck in lastCheckList:
            diskInfoObj = lastDiskCheckDic[lastCheck]
            _getDiskIdList().append(diskInfoObj.getDiskId())
    else:
        random.shuffle(_getDiskIdList())
    if (srvObj.getCfg().getDataCheckFileSeq() == NGAMS_CHECK_SEQ):
        _setSeqFileCheck(1)
    else:
        _setSeqFileCheck(0)
    info(4,"Ordered disks according to the sequence they should be checked")
    ###################################################################

    ###########################################################################
    # Initialize the statistics parameters for the checking.
    ###########################################################################
    info(4,"Initialize the statistics for the checking cycle ...")
    amountMb = 0.0
    noOfFiles = 0
    for diskId in _getDiskIdList():
        queueDbm = _getDbmObjDic()[diskId][0]
        queueDbm.initKeyPtr()
        while (1):
            fileKey, fileInfo = queueDbm.getNext(0)
            if (not fileKey): break
            noOfFiles += 1
            amountMb += float(float(fileInfo[ngamsDb.SUM1_FILE_SIZE]) /\
                              1048576.0)
    _initFileCheckStatus(srvObj, amountMb, noOfFiles)
    info(4,"Initialized the statistics for the checking cycle")
    ###########################################################################

    info(4,"Leaving _dumpFileInfo()")


def _schedNextFile(srvObj,
                   threadId):
    """
    Function that returns the information about the next file to be checked
    according to the checking scheme.

    The function keeps track of which disk has been allocated to which
    Checking Sub-Thread and allocates only files from the given disk to that
    thread.

    srvObj:       Reference to server object (ngamsServer).

    threadId:     ID of the sub-thread (string).
     
    Returns:      List with information (format see ngamsDb.getFileSummary1())
                  for file to be checked. If there are no more files to check,
                  None is returned (list/None).
    """
    info(4,"Entering _schedNextFile() ...")
    try:
        _takeReqFileInfoSem()
        fileInfo = None
        while (1):
            if (not _getDiskSchedDic().has_key(threadId)):
                if (_getDiskIdList()):
                    nextDiskId = _getDiskIdList()[0]
                    _getDiskSchedDic()[threadId] = nextDiskId
                    _getDbmObjDic()[nextDiskId][0].initKeyPtr()
                    del _getDiskIdList()[0]
                else:
                    break
            diskId = _getDiskSchedDic()[threadId]
            if (_getSeqFileCheck()):
                fileKey, fileInfo = _getDbmObjDic()[diskId][0].getNext(1)
                _getDbmObjDic()[diskId][0].sync()
            else:
                fileKey, fileInfo = _getDbmObjDic()[diskId][0].getRandom(1)
                _getDbmObjDic()[diskId][0].sync()
            if (fileKey):
                # We got a file key + file info list, return the info.
                break
            else:
                # There are no more file info lists for the given Disk ID,
                # remove that Disk ID from the list so that it switches to
                # the next. Also set, the time for the last check of that disk.
                _getDbmObjDic()[diskId][0].cleanUp()
                rmFile(_getDbmObjDic()[diskId][0].getDbmName())
                srvObj.getDb().setLastCheckDisk(diskId, time.time())
                if (_getDiskIdList()):
                    nextDiskId = _getDiskIdList()[0]
                    _getDiskSchedDic()[threadId] = nextDiskId
                    del _getDiskIdList()[0]
                    _getDbmObjDic()[nextDiskId][0].initKeyPtr()
                else:
                    break
        _relReqFileInfoSem()
        info(4,"Leaving _schedNextFile()")
        return fileInfo
    except Exception, e:
        _relReqFileInfoSem()
        raise e


def _dataCheckSubThread(srvObj,
                        threadId,
                        dummy):
    """
    Sub-thread scheduled to carry out the actual checking. This makes
    it possible to do the checking in several threads simultaneously if
    the NGAS Host has multiple CPUs.

    srvObj:       Reference to server object (ngamsServer).

    threadId:     ID allocated to this thread (string).

    Returns:      Void.
    """
    info(4,"Entering _dataCheckSubThread() ...")
    dataCheckPrio = srvObj.getCfg().getDataCheckPrio()
    while (1):
        _stopCheckingSubThread()

        try:
            # Get the info for the next file to check + check it.
            fileInfo = _schedNextFile(srvObj, threadId)
            if (not fileInfo):
                info(4,"No more files in queue to check - exiting")
                _updateFileCheckStatus(srvObj, None, None, None, None, [], 1)
                raise Exception, "_STOP_DATA_CHECK_SUB_THREAD_"

            # Remove the entry for this file in the File Reference DBM to
            # indicate that the file is registered in the DB.
            filename = os.path.normpath(fileInfo[ngamsDb.SUM1_MT_PT] + "/" +\
                                        fileInfo[ngamsDb.SUM1_FILENAME])
            if (_getFileRefDbm().hasKey(filename)):
                _getFileRefDbm().rem(filename).sync()

            # Update the overall status of the checking.
            tmpReport = []
            ngamsFileUtils.checkFile(srvObj, fileInfo, tmpReport,
                                     srvObj.getCfg().getDataCheckScan())
            if (not tmpReport): tmpReport = [[]]
            _updateFileCheckStatus(srvObj, fileInfo[ngamsDb.SUM1_FILE_SIZE],
                                   fileInfo[ngamsDb.SUM1_DISK_ID],
                                   fileInfo[ngamsDb.SUM1_FILE_ID],
                                   fileInfo[ngamsDb.SUM1_VERSION],
                                   tmpReport[0])

            # If the server is handling a command, the sub-thread will suspend
            # itself until the server is idle again.
            if (srvObj.getHandlingCmd()):
                while (srvObj.getHandlingCmd()):
                    _suspend(srvObj, 0.200)
        except Exception, e:
            if (str(e).find("_STOP_DATA_CHECK_SUB_THREAD_") != -1):
                thread.exit()
            else:
                error("Exception encountered in Data Check Sub-Thread: %s" %\
                      str(e))
                time.sleep(2)
            

def _genReport(srvObj):
    """
    Generate the DCC Check Report according to the problems found.

    srvObj:     Reference to instance of ngamsServer object (ngamsServer).

    Returns:    Void.
    """
    # Find out how many inconsistencies were found.
    noOfProbs = 0
    # Errors found.
    for diskId in _getDiskDic().keys():
        noOfProbs += _getDbmObjDic()[diskId][1].getCount()
    # Spurious files on disk.
    unRegFiles = _getFileRefDbm().getCount()

    # Generate the report.
    global _statCheckStart, _statCheckRate, _statCheckFiles, _statCheckedMb
    checkTime = (time.time() - _statCheckStart)
    if ((noOfProbs + unRegFiles) or srvObj.getCfg().getDataCheckForceNotif()):

        report    = ""
        hdrForm   = "%-20s %s\n"
        format    = "%-60s %-32s %-9s %s\n"
        separator = 130 * "-" + "\n"
       
        # Build up the report.
        report =  "DATA CHECKING REPORT:\n\n"
        report += hdrForm % ("Date", PccUtTime.TimeStamp().getTimeStamp())
        report += hdrForm % ("NGAS Host ID", getHostId())
        report += hdrForm % ("Start Time", PccUtTime.TimeStamp().\
                             initFromSecsSinceEpoch(_statCheckStart).\
                             getTimeStamp())
        report += hdrForm % ("Total Time (s)", "%.3f" % checkTime)
        report += hdrForm % ("Total Time (hours)", "%.3f" % (checkTime / 3600))
        report += hdrForm % ("Rate (MB/s)", "%.3f" % _statCheckRate)
        report += hdrForm % ("Files Checked", _statCheckFiles)
        report += hdrForm % ("Data Checked (MB)", "%.5f" % _statCheckedMb)
        report += hdrForm % ("Inconsistencies",  str(noOfProbs + unRegFiles))
        report += separator

        # Any descrepancies found?
        if ((not noOfProbs) and (not unRegFiles)):
            report += "No descrepancies found!"

        # Inconsistencies found?
        if (noOfProbs):
            report += "INCONSISTENT FILES FOUND:\n\n"
            report += format % ("Problem Description", "File ID",
                                "Version", "Slot ID:Disk ID")
            report += separator
            for diskId in _getDiskDic().keys():
                errDbm = _getDbmObjDic()[diskId][1].initKeyPtr()
                while (1):
                    fileKey, errInfo = errDbm.getNext(0)
                    if (not fileKey): break
                    slotDiskId = errInfo[3] + ":" + errInfo[4]
                    report += format % (errInfo[0], errInfo[1], errInfo[2],
                                        slotDiskId)
            report += separator

        # Not registered files found?
        if (unRegFiles):
            repFormat = "%-32s %s\n"
            report += "NON REGISTERED FILES FOUND ON STORAGE DISKS:\n\n"
            report += repFormat % ("Disk ID:", "Filename:")
            report += separator
            _getFileRefDbm().initKeyPtr()
            while (1):
                filename, fileInfo = _getFileRefDbm().getNext(0)
                if (not filename): break
                report += repFormat % (fileInfo[0], filename)
            report += separator
            
        # Send Notification Message if needed (only if disks where checked).
        if (len(_getDiskDic().keys())):
            ngamsNotification.notify(srvObj.getCfg(), NGAMS_NOTIF_DATA_CHECK,
                                     "DATA CHECK REPORT", report, [], 1)

    # Give out the statistics for the checking.
    if (srvObj.getCfg().getDataCheckLogSummary()):
        msg = genLog("NGAMS_INFO_DATA_CHK_STAT",
                     [_statCheckFiles,_statCheckedMb,_statCheckRate,checkTime])
        info(0,msg)
        
    # Remove the various DBMs allocated.
    for diskId in _getDiskDic().keys():
        _getDbmObjDic()[diskId][1].cleanUp()
        del _getDbmObjDic()[diskId]
    _resetDbmObjDic()
    _getFileRefDbm().cleanUp()
    _resetFileRefDbm()
    

def dataCheckThread(srvObj,
                    dummy):
    """
    The Data Check Thread is executed to run a periodic check of the
    consistency of the data files contained in an NG/AMS system. The periodic
    check is performed only when NG/AMS is Online.

    srvObj:       Reference to server object (ngamsServer).
     
    Returns:      Void.
    """
    minCycleTime=ngamsLib.isoTime2Secs(srvObj.getCfg().getDataCheckMinCycle())
    dataCheckThreadLoop = 1
    while (dataCheckThreadLoop):

        # Incapsulate this whole block to avoid that the thread dies in
        # case a problem occurs, like e.g. a problem with the DB connection.
        try:
            # Wait until we're sure that the Janitor Thread has executed
            # at least once, to ensure that the check is carried out in a
            # clean environment.
            while (not srvObj.getJanitorThreadRunCount()):
                _stopDataCheckThr(srvObj)
                time.sleep(0.5)

            # Ensure that DB connection is active.
            srvObj.getDb().checkCon()

            if (srvObj.getCfg().getDataCheckLogSummary()):
                info(0,"Data Check Thread starting iteration ...")

            srvObj.updateHostInfo(None, None, None, None, None, None, 1, None)

            # It was intended to implement a scheme for preserving the info
            # about files already checked between two sessions if the DCC
            # should be interrupted if the server is shut down. This attempt
            # has been given up since if the disk configuration changed
            # between the two sessions, false DCC problems might be reported.
            #
            # These temporary DCC DBM's are kept in the NG/AMS Cache Dir
            # since if put in the NG/AMS Temp. Dir. they might be deleted
            # if not accessed for the specified expiration time in the temp.
            # dir - a DCC run may run for a longer period of time.
            rmFile("%s/%s/%s*" % (srvObj.getCfg().getMountRootDirectory(),
                                  NGAMS_CACHE_DIR, NGAMS_DATA_CHECK_THR))
 
            # Get the information about the files to check.
            try:
                tmpFilePat = ngamsHighLevelLib.\
                             genTmpFilename(srvObj.getCfg(),
                                            NGAMS_DATA_CHECK_THR)
                _dumpFileInfo(srvObj, tmpFilePat)
                rmFile(tmpFilePat + "*")
            except Exception, e:
                rmFile(tmpFilePat + "*")
                raise e
 
            # According to the number of disks to be checked, a sub-thread
            # is allocated for each up to the limit defined in the
            # configuration.
            #
            # Afterwards the main DCC Thread monitors the execution of
            # these + update the status information.
            thrHandleDic = {}
            noOfSubThreads = len(_getDiskIdList())
            if (noOfSubThreads > srvObj.getCfg().getDataCheckMaxProcs()):
                noOfSubThreads = srvObj.getCfg().getDataCheckMaxProcs()
            _resetDiskSchedDic()
            for n in range (1, (noOfSubThreads + 1)):
                threadId = NGAMS_DATA_CHECK_THR + "-" + str(n)
                args = (srvObj, threadId, None)
                info(4,"Starting Data Check Sub-Thread: %s" % threadId)
                thrHandleDic[n] = threading.Thread(None, _dataCheckSubThread,
                                                   threadId, args)
                thrHandleDic[n].setDaemon(0)
                thrHandleDic[n].start()
            while (1):
                time.sleep(0.200)

                # Check if the DCC Thread is allowed to run, else stop all
                # activities.
                if (not srvObj._dataCheckRunSync.isSet()):
                    # Stop the sub-threads.
                    _setRunPermSubThread(0)
                    startTime = time.time()
                    noOfSubThr = len(thrHandleDic.keys())
                    while ((time.time() - startTime) < 10):
                        thrCount = 0
                        for thrId in thrHandleDic.keys():
                            if (not thrHandleDic[thrId].isAlive()):
                                del thrHandleDic[thrId]
                                thrCount += 1
                        if (thrCount == noOfSubThr): break
                        time.sleep(0.050)
                    # Kill the possible remaining sub-threads.
                    for thrId in thrHandleDic.keys():
                        del thrHandleDic[thrId]
                    _stopDataCheckThr(srvObj)
                else:
                    # Check if all the sub-threads are still running
                    # or if the check cycle is completed.
                    for thrId in thrHandleDic.keys():
                        if (not thrHandleDic[thrId].isAlive()):
                            del thrHandleDic[thrId]
                    if (not len(thrHandleDic)):
                        lastCheckTime = time.time()
                        break
            
            # Send out check report if any discrepancies found + send
            # out notification message according to configuration.
            _genReport(srvObj)

            # Set the last check for all disks to the same value
            for diskId in _getDiskDic().keys():
                srvObj.getDb().setLastCheckDisk(diskId, lastCheckTime)

            srvObj.updateHostInfo(None, None, None, None, None, None, 0, None)
            ###################################################################

            ###################################################################
            # Check if we should wait for a while for the Minimum Cycle
            # Time to elapse.
            ###################################################################
            lastOldestCheck = srvObj.getDb().getMinLastDiskCheck(getHostId())
            waitTime = 0
            timeNow = time.time()
            if (lastOldestCheck):
                if ((timeNow - lastOldestCheck) < minCycleTime):
                    waitTime = (minCycleTime - (timeNow - lastOldestCheck))
            else:
                execTime = (timeNow - _statCheckStart)
                if (execTime < minCycleTime):
                    waitTime = (minCycleTime - execTime)
            if (waitTime):
                waitTime += 1
                info(4,"Suspending Data Checking Thread for: " +\
                     str(waitTime)+"s.")
                nextAbsCheckTime = int(time.time() + waitTime + 0.5)
                info(3,"Next Data Checking scheduled for: " +\
                     PccUtTime.TimeStamp().\
                     initFromSecsSinceEpoch(nextAbsCheckTime).\
                     getTimeStamp() + "/" + str(nextAbsCheckTime))
                srvObj.setNextDataCheckTime(nextAbsCheckTime)            
                tmpTimer = PccUtTime.Timer()
                run = 1
                while (run):
                    _stopDataCheckThr(srvObj)
                    tmpTimeNow = tmpTimer.getLap()
                    if ((waitTime - tmpTimeNow) >= 2):
                        sleepTime = 2
                    else:
                        sleepTime = (waitTime - tmpTimeNow)
                        run = 0
                    time.sleep(sleepTime)
            ###################################################################
            
        except Exception, e:
            if (str(e).find("_STOP_DATA_CHECK_THREAD_") != -1):
                thread.exit()
            elif (str(e).find("NGAMS_ER_DB_COM") != -1):
                if (getTestMode()):
                    waitTime = 1
                else:
                    waitTime = 3600
                errMsg = "Problem communicating with DB. " +\
                         "Suspending DCC %ds. Error : %s"
                errMsg = errMsg % (waitTime, str(e))
                error(errMsg)
                ngamsNotification.notify(srvObj.getCfg(),
                                         NGAMS_NOTIF_DATA_CHECK,
                                         "DATA CHECK ERROR", errMsg, [], 1)
                startWaitTime = time.time()
                while ((time.time() - startWaitTime) < waitTime):
                    if (not srvObj._dataCheckRunSync.isSet()):
                        info(2,"Stopping Data Check Thread")
                        srvObj.updateHostInfo(None, None, None, None, None,
                                              None, 0, None)
                        srvObj._dataCheckRunSync.set()
                        thread.exit()
                    time.sleep(2)
            else:
                errMsg = "Error occurred during execution of the " +\
                         "Data Check Thread. Exception: " + str(e)
                error(errMsg)
                # We make a small wait here to avoid that the process tries
                # too often to carry out the tasks that failed.
                time.sleep(1.0)


# EOF
