#******************************************************************************
# ESO/DFS
#
# "@(#) $Id: ngamsStatusCmd.py,v 1.44 2005/12/12 10:20:35 jknudstr Exp $"
#
# Who       When        What
# --------  ----------  -------------------------------------------------------
# jknudstr  22/04/2003  Created
#

"""
Function + code to handle the STATUS command.
"""
import time, sys, types

import pcc, PccUtTime
from ngams import *
import ngamsDb, ngamsLib, ngamsStatus, ngamsDiskInfo
import ngamsFileInfo, ngamsHighLevelLib, ngamsFileUtils
import ngamsSrvUtils

# Man-page for the command.
_help = loadDoc("ngamsServer/ngamsStatusCmd.doc")


def _checkFileAccess(srvObj,
                     reqPropsObj,
                     httpRef,
                     fileId,
                     fileVersion = -1,
                     diskId = ""):
    """
    Check if a file is accessible either on the local host or on a remotely
    located host.

    srvObj:         Instance of the server object (ngamsServer).

    reqPropsObj:    Request Property object to keep track of actions done
                    during the request handling (ngamsReqProps).

    httpRef:        Reference to the HTTP request handler
                    object (ngamsHttpRequestHandler).
    
    fileId:         File ID (string).
    
    fileVersion:    File Version (integer).
    
    diskId:         Disk ID (string).

    Returns:        Returns message indicating if the file is available
                    (string).
    """
    info(4,"Checking for access to file with ID: " + fileId + " ...")

    # Check if the file is located on this host, or if the request should be
    # forwarded (if this server should act as proxy).
    location, fileHost, ipAddress, filePortNo, mountPoint, filename, fileId,\
              fileVersion, mimeType =\
              ngamsFileUtils.locateArchiveFile(srvObj, fileId, fileVersion,
                                               diskId)
    if (location != NGAMS_HOST_LOCAL):
        httpStatCode, httpStatMsg, httpHdrs, data =\
                      srvObj.forwardRequest(reqPropsObj, httpRef, fileHost,
                                            filePortNo, autoReply = 0)
        tmpStat = ngamsStatus.ngamsStatus().unpackXmlDoc(data)
        return tmpStat.getMessage()
    else:
        # First check if this system allows for Retrieve Requests.
        if (not srvObj.getCfg().getAllowRetrieveReq()):
            msg = genLog("NGAMS_INFO_SERVICE_UNAVAIL", ["File Retrieval"])
        else:
            fileHost = "%s:%d" % (getHostName(), filePortNo)
            msg = genLog("NGAMS_INFO_FILE_AVAIL",
                         [fileId + "/Version: " + str(fileVersion), fileHost])
        return msg


def _getRefCounts():
    """
    Return information about all object allocated and the number of
    references pointing to each object. This can be used to check a running
    server for memory (object) leaks.

    Taken from: http://www.nightmare.com/medusa/memory-leaks.html.

    Returns:    
    """
    d = {}
    sys.modules
    # Collect all classes
    for m in sys.modules.values():
        for sym in dir(m):
            o = getattr (m, sym)
            if type(o) is types.ClassType:
                d[o] = sys.getrefcount (o)
    # sort by refcount
    pairs = map(lambda x: (x[1],x[0]), d.items())
    pairs.sort()
    pairs.reverse()
    return pairs


def _genRefCountRep():
    """
    Generate a report with number of objects allocated of each type.
    The report is a list with a sublist for each object of the form:

      [[<No of Refs>, <Object>], ...]

    Returns:   Report (list).
    """
    refDic = {}
    for count, obj in _getRefCounts():
        if (not refDic.has_key(obj)): refDic[obj] = 0
        refDic[obj] += count
    objList = []
    objects = refDic.keys()
    objects.sort()
    for obj in objects:
        objList.append([refDic[obj], obj])
    return objList

  
def _genObjectStatus():
    """
    Generate a report with information about objects allocated, numbers of
    references to each object and other information that can be used to
    track down memory (object) leaks in the server.
    
    Returns:  Object report (string).
    """
    rep = "NG/AMS SERVER OBJECT STATUS REPORT\n\n"
    import gc
    gc.set_debug(gc.DEBUG_COLLECTABLE | gc.DEBUG_UNCOLLECTABLE |
                 gc.DEBUG_INSTANCES | gc.DEBUG_OBJECTS)
    rep += "=Garbage Collector Status:\n\n"
    rep += "Enabled:             %d\n" % gc.isenabled()
    rep += "Unreachable objects: %d\n" % gc.collect()
    rep += "Threshold:           %s\n" % str(gc.get_threshold())
    #rep += "Objects:\n"
    #rep += str(gc.get_objects()) + "\n"
    rep += "Garbage:\n"
    rep += str(gc.garbage) + "\n\n"

    # Dump refence count status of all objects allocated into file specified.
    rep += "=Object Reference Counts:\n\n"
    for objInfo in _genRefCountRep():
        rep += "%-4d %s\n" % (objInfo[0], objInfo[1])
        
    rep += "\n=EOF"
    return rep

        
def handleCmdStatus(srvObj,
                    reqPropsObj,
                    httpRef):
    """
    Handle STATUS command.
        
    srvObj:         Reference to NG/AMS server class object (ngamsServer).
    
    reqPropsObj:    Request Property object to keep track of
                    actions done during the request handling
                    (ngamsReqProps).
        
    httpRef:        Reference to the HTTP request handler
                    object (ngamsHttpRequestHandler).
        
    Returns:        Void.
    """
    info(4,"Entering handleCmdStatus() ...")
    status = ngamsStatus.ngamsStatus()
    status.\
             setDate(PccUtTime.TimeStamp().getTimeStamp()).\
             setVersion(getNgamsVersion()).setHostId(getHostId()).\
             setStatus(NGAMS_SUCCESS).\
             setMessage("Successfully handled command STATUS").\
             setState(srvObj.getState()).setSubState(srvObj.getSubState())

    reqPropsObjRef = reqPropsObj
        
    # Get the information requested.
    diskId            = ""
    fileId            = ""
    fileVersion       = ""
    configurationFile = ""
    fileAccess        = ""
    hostId            = ""
    requestId         = ""
    if (reqPropsObj.hasHttpPar("disk_id")):
        diskId = reqPropsObj.getHttpPar("disk_id")
    if (reqPropsObj.hasHttpPar("file_id")):
        fileId = reqPropsObj.getHttpPar("file_id")
    if (reqPropsObj.hasHttpPar("file_version")):
        fileVersion = reqPropsObj.getHttpPar("file_version")
    if (reqPropsObj.hasHttpPar("configuration_file")): configurationFile = "-"
    if (reqPropsObj.hasHttpPar("file_access")):
        fileAccess = reqPropsObj.getHttpPar("file_access")
    if (reqPropsObj.hasHttpPar("host_id")):
        hostId = reqPropsObj.getHttpPar("host_id")
    if (reqPropsObj.hasHttpPar("request_id")):
        requestId = reqPropsObj.getHttpPar("request_id")
    if (reqPropsObj.hasHttpPar("flush_log")):
        try:
            logFlush()
        except:
            pass
    if (reqPropsObj.hasHttpPar("dump_object_info")):
        # Dump status of all objects allocated into file specified.
        targFile = reqPropsObj.getHttpPar("dump_object_info")
        rmFile(targFile)
        fo = open(targFile, "w")
        fo.write(_genObjectStatus())
        fo.close()

    # Handle request for file info.
    genCfgStatus     = 0
    genDiskStatus    = 0
    genFileStatus    = 0
    genStatesStatus  = 1
    genRequestStatus = 0
    msg              = ""
    help             = 0
    if (reqPropsObj.hasHttpPar("help")):
        global _help
        msg = _help
        help = 1
    elif (hostId and (hostId != getHostId())):
        # Query the status for the host referenced.
        contactDic = ngamsHighLevelLib.\
                     resolveHostAddress(srvObj.getDb(), srvObj.getCfg(),
                                        [hostId])
        # Handle the request as follows:
        #
        # 1. Resolved host/port = local host/port:
        #    => Generate standard reply to STATUS command.
        #
        # 2. Resolved host/port != local host/port + !Proxy Mode:
        #    => Send back Re-Direction HTTP Response.
        #
        # 3. Resolved host/port != local host/port + Proxy Mode:
        #    => Forward the request to the host indicated, and send
        #       back the reply received from this.
        hostObj = contactDic[hostId]
        cfgObj = srvObj.getCfg()
        if ((hostObj.getHostId() == getHostId()) and 
            (hostObj.getSrvPort() == cfgObj.getPortNo())):
            info(2,"Send back status of this server/host to STATUS/host_id "+\
                 "request")
            msg = "Successfully handled command STATUS"
        elif (((hostObj.getHostId() != getHostId()) or
               (hostObj.getSrvPort() != cfgObj.getPortNo())) and
              (not cfgObj.getProxyMode())):
            info(2,"Sending back re-direction HTTP response for host/port: "+\
                 "%s/%d to STATUS/host_id request" %
                 (hostObj.getHostId(), hostObj.getSrvPort()))
            srvObj.httpRedirReply(reqPropsObj, httpRef, hostObj.getHostId(),
                                  hostObj.getSrvPort())
            return
        else:
            try:
                # Check if host is suspended, if yes, wake it up.
                if (srvObj.getDb().getSrvSuspended(hostObj.getHostId())):
                    info(3,"Status Request - Waking up suspended " +\
                         "NGAS Host: " + hostObj.getHostId())
                    ngamsSrvUtils.wakeUpHost(srvObj, hostObj.getHostId())
                srvObj.forwardRequest(reqPropsObj, httpRef,hostObj.getHostId(),
                                      hostObj.getSrvPort(), autoReply = 1)
            except Exception, e:
                ex = re.sub("<|>", "", str(e))
                errMsg = genLog("NGAMS_ER_COM",
                                [hostObj.getHostId(),hostObj.getSrvPort(),ex])
                raise Exception, errMsg
            return
    elif (diskId):
        diskObj = ngamsDiskInfo.ngamsDiskInfo()
        diskObj.read(srvObj.getDb(), diskId)
        status.addDiskStatus(diskObj)
        genDiskStatus = 1
    elif (fileId):
        if (not fileVersion): fileVersion = 1
        fileObj = ngamsFileInfo.ngamsFileInfo()
        fileObj.read(srvObj.getDb(), fileId, fileVersion)
        diskObj = ngamsDiskInfo.ngamsDiskInfo()
        try:
            diskObj.read(srvObj.getDb(), fileObj.getDiskId())
        except:
            errMsg = "Illegal Disk ID found: %s for file with ID: %s" %\
                     (fileObj.getDiskId(), fileId)
            raise Exception, errMsg
        diskObj.addFileObj(fileObj)
        status.addDiskStatus(diskObj)
        genDiskStatus = 1
        genFileStatus = 1
    elif (requestId):
        info(3,"Checking status of request with ID: " + requestId)
        reqPropsObjRef = srvObj.getRequest(requestId)
        if (not reqPropsObjRef):
            errMsg = genLog("NGAMS_ER_ILL_REQ_ID", [requestId])
            raise Exception, errMsg
        genRequestStatus = 1
    elif (configurationFile):
        msg = "configuration_file=" + srvObj.getCfg().getCfg()
        genCfgStatus = 1
        status.setNgamsCfgObj(srvObj.getCfg())
    elif (fileAccess):
        if (not fileVersion): fileVersion = -1
        fileId = fileAccess
        msg = _checkFileAccess(srvObj, reqPropsObj, httpRef, fileId,
                               fileVersion, diskId)
    else:
        msg = "Successfully handled command STATUS"

    if (reqPropsObjRef == reqPropsObj):
        reqPropsObj.setCompletionTime()
        srvObj.updateRequestDb(reqPropsObj)
    if (genCfgStatus or genDiskStatus or genFileStatus or genRequestStatus):
        status.setReqStatFromReqPropsObj(reqPropsObjRef)
        
        # Generate XML reply.
        xmlStat = status.genXmlDoc(genCfgStatus, genDiskStatus, genFileStatus,
                                   genStatesStatus)
        xmlStat = ngamsHighLevelLib.\
                  addDocTypeXmlDoc(srvObj, xmlStat, NGAMS_XML_STATUS_ROOT_EL,
                                   NGAMS_XML_STATUS_DTD)
        srvObj.httpReply(reqPropsObj, httpRef, NGAMS_HTTP_SUCCESS, xmlStat,
                         NGAMS_XML_MT)
    else:
        srvObj.reply(reqPropsObj,httpRef,NGAMS_HTTP_SUCCESS,NGAMS_SUCCESS,msg)
        
    info(4,"Leaving handleCmdStatus()")
    if (msg and (not help)):
        info(1,msg)
    else:
        info(1,"Successfully handled command STATUS")


# EOF
