#!/usr/bin/python
#******************************************************************************
# ESO/DMD
#
# "@(#) $Id: vultur_exec_cascade.py,v 1.3 2006/03/13 14:55:27 jknudstr Exp $"
#
# Who       When        What
# --------  ----------  -------------------------------------------------------
# jknudstr  01/04/2005  Created
#

"""
The Vultur Cascade Execution Tool is used to facilitate the submissions
of DAGs (cascades) to the a Condor Pool. The user does not have to worry
about the low-level preparation for running the DAG and does not have
to provide all the necessary files for executing the DAG.

SYNOPSIS:
The command line options for the tool are:

  $ vultur_exec_cascade --dagId=<ID> --jobs=<Job List File> [--wait] \
    [--tmpDir=<Job Tmp Dir>]

The options are:

  --dagId:       Overall ID for the DAG.

  --jobs:        Reference to the DAG Submission File containing the definition
                 of all the jobs in the DAG. The format is explained in more
                 details below. 

  --wait:        Make the tool wait (block) until the DAG has finished
                 executing.

  --tmpDir:      Under the given path, a temporary working directory per job
                 will be created. This directory will be named:

                   <Job Tmp Dir>/<Job ID>

                 From this directory the job will be executed. If this
                 parameter is not given, the Job Working Directory will be
                 created within the DAG Working Directory.
  

DAG SUBMISSION FILE:
The DAG Submission file is a simplified interface for the user to specify
a set of jobs and their properties. The format makes it possible to specify
the jobs of a DAG in a relatively compact format.

The format is:

<Job ID>||<Executable>||<Args>||<Dependencies>||<Upload Files>

The fields are as follows:

Job ID:          ID of the job. Used to make references in the DAG Submission
                 File when setting up dependencies.

Executable:      Executable to execute on the Execute Node.

Args:            Arguments to pass on to the executable.

Dependencies:    Comma separated list of Jobs IDs, on which this job depends.
                 This job will first execute when the jobs specified by the
                 Job IDs have finished execution. 

Upload Files:    Comma separated list of files to upload to the Execute Node
                 when the DAG executes.


The fields are separated by a double pipe character to allow to use '|'
as arguments to the executable of the job.


EXECUTION ENVIRONMENT:
The tool prepares to run the DAG in the following way:

1. It creates a temporary directory, which will host the files, pertinent
to the DAG. This will be located in:

  /data/<User>/condor/<DAG Tmp Dir>

- where <DAG ID> is the ID of the DAG allocated by the user. The tool will
concatenate a timestamp to the ID to make the directory unique.

2. In this directory the following files will be contained:

  a. Job Submission File for the individual job (<Job ID>.cmd).
  b. Log files for each job (<Job ID>.err/.log/.out).
  c. The DAG Submission file (<DAG ID>.dag).
  d. The Condor DAG handling files (<DAG ID>.dag.*.sub/.log/.out/.rescue).
  e. The DOT files used to visualize the state transitions of the DAG
     (<DAG ID>.dot.<#>).
  f. If the -tmpDir parameter is not specified, the Job Working Directories
     will be contained in this directory as well.

The temporary directory is visible and accessible from all nodes in a pool
configuration.

Note, no attempts are done to clean up the temporary directories automatically.
This can be done by invoking the tool vultur_clean_up. This could be invoked
periodically (system wide) via a cron job if desirable.
"""

import sys, os, time, commands
from vultur_utils import *


def correctUsage():
    """
    Print tool synopsis and exit.

    Returns:   Void.
    """
    print "\nCorrect usage is:\n\n" +\
          "$ vultur_exec_cascade --dagId=<ID> --jobs=<Job List File> " +\
          "[--wait] [--tmpDir=<Root Job Work Dir>] \n\n" +\
          "MANUAL PAGE:" + __doc__ + "\n"

def submitDag(argv):
    """
    Parse command line parameters and submit the DAG.

    argv:  Array of command line parameters (list).

    Returns:   Void.
    """
    dagId = jobListFile = wait = tmpDir = None
    idx = 1
    while idx < len(sys.argv):
        par = sys.argv[idx].upper()
        if (par.find("--DAGID") == 0):
            dagId = sys.argv[idx].split("=")[1].strip()
        elif (par.find("--JOBS") == 0):
            jobListFile = sys.argv[idx].split("=")[1].strip()
        elif (par.find("--WAIT") == 0):
            wait = 1
        elif (par.find("--TMPDIR") == 0):
            tmpDir = sys.argv[idx].split("=")[1].strip()
        elif ((par == "-H") or (par.find("--HELP") == 0)):
            correctUsage()
            sys.exit(0)
        else:
            correctUsage()
            sys.exit(1)
        idx += 1

    if ((not dagId) or (not jobListFile)):
        correctUsage()
        sys.exit(1)
    user = os.environ["USER"]

    # Load in job list.
    fo = open(jobListFile)
    tmpJobListFile = fo.readlines()
    fo.close()
    jobList = []
    for line in tmpJobListFile:
        line = line.strip()
        if (line == ""): continue
        if (line[0] == "#"): continue
        els = line.split("||")
        subJobId, appl, args = els[0], els[1], els[2]
        if (len(els) > 3):
            parIds = els[3].strip()
        else:
            parIds = None
        jobList.append((subJobId.strip(), appl.strip(),
                        args.strip(), parIds))

    # Create directory from within which DAG is launched (will host all files
    # generated by the DAG execution).
    dagBaseDir = os.path.normpath("%s/condor/%s-%.8f" %\
                                  (os.environ["BQSROOT"], dagId, time.time()))
    os.system("mkdir -p %s" % dagBaseDir)
    info("DAG Base Directory is: %s" % dagBaseDir)
    dagFile = "%s/%s.dag" % (dagBaseDir, dagId)
    dagFile = os.path.normpath(dagFile)

    # Generate Job Submission Files.
    if (tmpDir):
        jobRootWorkDir = tmpDir
    else:
        jobRootWorkDir = dagBaseDir
    jobSubmisList = []
    for job in jobList:
        subJobId, jobAppl, jobArgs, parIds = job
        jobWorkDir = os.path.normpath("%s/%s" % (jobRootWorkDir, subJobId))
        os.system("mkdir -p %s" % jobWorkDir)
        jobSubmisName, jobSubmis = genJobSubmission(subJobId, user, jobAppl,
                                                    jobArgs, 1, dagBaseDir,
                                                    jobWorkDir)
        jobSubmisList.append((subJobId, jobSubmisName, parIds))
    
    # Build DAG.
    dag = ""
    # Define jobs.
    for job in jobSubmisList:
        dag += "Job %16s %s\n" % (job[0], job[1])
    dag += "\n"

    # Set up job dependencies.
    dependLine = "PARENT %s CHILD %s\n"
    for job in jobSubmisList:
        if (job[2].find("NONE") != -1): continue
        dag += dependLine % (job[2].replace(",", " ").strip(), job[0])

    # Request DOT files to be produced.
    dag += "\nDOT %s.dot DONT-OVERWRITE UPDATE\n\n" % dagId

    # Finish up DAG + write it. 
    dag += "# EOF\n"
    os.system("rm -f %s" % dagFile)
    fo = open(dagFile, "w")
    fo.write(dag)
    fo.close()
    info("Wrote DAG: %s" % dagFile)

    # Execute DAG.
    # - Enter the DAG directory.
    dagStart = time.time()
    dagExecCmd = "cd %s; condor_submit_dag -f -notification Never %s" %\
                 (dagBaseDir, os.path.basename(dagFile))
    info("Submitting DAG with command: %s" % dagExecCmd)
    stat, out = commands.getstatusoutput(dagExecCmd)
    info("Status of DAG submission: %d/%s" % (stat, out))

    # Wait for the cascade to finish if requested.
    if (wait):
        # Get the Condor cluster ID for the DAGman job.
        dagManId = out.split("job(s) submitted to cluster ")[-1].\
                   split(".")[0] + ".0"
        sys.stdout.write("Waiting for DAG: %s to finish execution " %\
                         dagId)
        sys.stdout.flush()
        # Enter a loop and poll for the job to finish.
        count = 0
        while (1):
            if ((count % 30) == 0):
                sys.stdout.write(".")
                sys.stdout.flush()
            if ((count % 5) == 0):
                statCmd = "condor_q %s %s" % (os.environ["USER"], dagManId)
                stat, out = commands.getstatusoutput(statCmd)
                statLines = out.split("\n")
                running = 0
                for statLine in statLines:
                    if (statLine.strip().find(dagManId) == 0):
                        running = 1
                        break
                if (not running):
                    dagStop = time.time()
                    msg = "\nDAG with ID: %s finished execution. " +\
                          "Total time: %.3fs"
                    print msg % (dagId, (dagStop - dagStart))
                    break
            time.sleep(1)
            count += 1
    

if __name__ == '__main__':
    """
    Main function.
    """
    submitDag(sys.argv)
    sys.exit(0)

# EOF
